﻿{-# LANGUAGE KindSignatures #-}
{-# LANGUAGE Rank2Types #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE DeriveDataTypeable #-}
{-# LANGUAGE TupleSections #-}

module PView (PView (..),

			  Source (..),
			  Lens (..),
			  ParametricLens (..),

			  read, 
			  update, 
			  observe) where

import Control.Monad
import Control.Monad.Trans
import Control.Monad.State (StateT, execStateT, evalStateT)
import Control.Monad.Writer (WriterT, execWriterT, tell)
import qualified Control.Monad.State as ST (get, gets, modify)
import Control.Applicative

import Data.Typeable
import Data.Dynamic

import Data.List
import Data.Maybe

import Prelude hiding (read)

-------------------------------------------------------------------------------

instance Show (NRequest m) where
   show (NRequest rid _ dyn) = "\n" ++ show rid ++ ": " ++ show (dynTypeRep dyn)

instance Show (NEvent m) where
   show (NEvent vid _) = "\n" ++ show vid

whenM :: (Applicative m, Monad m) => m Bool -> m () -> m ()
whenM f ift = f >>= \b -> if b then ift else return ()

ifM :: (Applicative m, Monad m) => m Bool -> m a -> m a -> m a
ifM f ift iff = f >>= \b -> if b then ift else iff

-------------------------------------------------------------------------------

-- Monad transformer for the notification engine
type PViewT m = StateT [NRequest m] m
  
-- Invalidation function: pure and monadic (utilized in the engine)
type Invalidate phi    = phi -> Bool
type InvalidateM m phi = phi -> PViewT m Bool

-- Notification request/event
type ObserveId = String
type ObserveHnd m = m ()

data NEvent m = forall phi. Typeable phi => NEvent TypeRep (InvalidateM m phi)
data NRequest m = NRequest ObserveId (ObserveHnd m) Dynamic

-- Parametric view GADT   
data PView phi m r w where
    Source      :: (Applicative m, Monad m, Typeable phi) 
                => Source phi m r w                          -> PView phi m r w
    
    Project     :: (Applicative m, Monad m, Typeable phi)
                => PView phi m r w -> Lens r w r' w'         -> PView phi m r' w'  
    
    Translate   :: (Applicative m, Monad m, Typeable phi)
                => PView phi m r w -> (psi -> phi)           -> PView psi m r w 
    
    Focus       :: (Applicative m, Monad m, Typeable phi, Typeable psi, Eq phi)
                => PView phi m r w -> ParametricLens psi r w -> PView (phi, psi) m r w 
    
    Product     :: (Applicative m, Monad m, Typeable phi, Typeable psi)
                => PView phi m r w -> PView psi m r' w'      -> PView (phi, psi) m (r, r') (w, w') 
    
    Merge       :: (Applicative m, Monad m, Typeable phi, Typeable psi)
                => PView phi m r w -> PView psi m r w   
                -> (phi -> r -> w -> Invalidate psi) 
                -> (psi -> r -> w -> Invalidate phi)              -> PView (Either phi psi) m r w

    Compose     :: (Applicative m, Monad m, Typeable phi, Typeable a)
                => PView phi m a a -> PView a m r w          -> PView (Composed phi a) m (a, r) (a, w) 

data Composed phi psi = MkComposed phi
                
data Source phi m r w where
    MkSource :: (Applicative m, Monad m) => (phi -> m r) -> (phi -> w -> m (Invalidate phi)) -> Source phi m r w
                
data Lens r w r' w' = MkLens { 
  get  :: r       -> r',
  put  :: r -> w' -> w
}

data ParametricLens psi r w = MkPLens {
  getf  :: psi -> r -> r,
  putf  :: psi -> r -> w -> (w, Invalidate psi)
}

-- Reading a parametric view
read :: (Applicative m, Monad m) => PView phi m r w -> phi -> PViewT m r
read (Source (MkSource sread _)) p = lift (sread p)
read (Project v l) p = get l <$> read v p
read (Translate v tr) p = read v (tr p)
read (Focus v l) (p,q) = getf l q <$> read v p
read (Product vl vr) (p,q) = (,) <$> read vl p <*> read vr q
read (Merge vl _ _ _) (Left p) = read vl p
read (Merge _ vr _ _) (Right p) = read vr p
read (Compose viewf views) (MkComposed p) = read viewf p >>= \a -> read views a >>= \b -> return (a,b)

-- Update a parametric view:
-- 1. do the actual update while generating notification events
-- 2. match notification events with notification requests
update :: (Applicative m, Monad m, Typeable phi) => PView phi m r w -> phi -> w -> PViewT m ()
update v p w = execWriterT (update' v p w) >>= trigger . reverse

trigger :: (Applicative m, Monad m) => [NEvent m] -> PViewT m ()
trigger es = do
  rs <- ST.get
  foldM_ (\skips event -> foldM (match event) skips rs) [] es
  where
    match (NEvent vid inval) skips (NRequest oid ohnd dyn)
      | notElem oid skips && dynTypeRep dyn == vid
         = case fromDynamic dyn of
             Just p -> whenM (inval p) (lift ohnd) >> return (oid:skips)
             Nothing  -> return skips
    match _ skips _ = return skips

returnE p inval = tell [NEvent (typeOf p) inval] >> return inval
         
update' :: (Applicative m, Monad m, Typeable phi) => PView phi m a b -> phi -> b -> WriterT [NEvent m] (PViewT m) (InvalidateM m phi)
update' (Source (MkSource _ supdate)) p w 
  = (lift.lift) (supdate p w) >>= \inval -> returnE p (return . inval)
              
update' (Project v l) p w 
  = lift (read v p) >>= \a -> update' v p (put l a w)

update' (Translate v tr) p w = (.tr) <$> update' v (tr p) w

update' (Focus v l) (p,q) w = do
  a <- lift $ read v p
  let (w', invall) = putf l q a w
  invalw <- update' v p w'
  returnE (p,q) (comp invall invalw)
  where 
    comp invall invalw (p',q')
      | p == p'   = return . invall $ q'
      | otherwise = invalw p'
    
update' (Product vl vr) (p,q) (wl,wr) = do
  invall  <- update' vl p wl
  invalr  <- update' vr q wr
  return  $ \(p,q) -> (||) <$> invall p <*> invalr q
    
update' (Merge v _ invall _) pp@(Left p) w = do
  r <- lift $ read v p 
  invalw <- update' v p w
  returnE pp (either invalw (return . invall p r w))    

update' (Merge _ v _ invalr) pp@(Right p) w = do
  r <- lift $ read v p 
  invalw <- update' v p w
  returnE pp (either (return . invalr p r w) invalw)    
    
update' (Compose vg vf) (MkComposed p) (wg,wf) = do
  invalg <- update' vg p wg
  invalf <- update' vf wg wf
  return $ \(MkComposed p) -> (||) <$> invalg p <*> (read vg p >>= invalf)

-- Subscribing for change notification
observe :: (Applicative m, Monad m, Typeable phi) => PView phi m r w -> phi -> ObserveId -> ObserveHnd m -> PViewT m ()
observe v p oid ohnd = do
  rs <- genreqs v p
  ST.modify (map (NRequest oid ohnd) rs ++)
  where
    genreqs :: (Applicative m, Monad m, Typeable phi) => PView phi m a b -> phi -> PViewT m [Dynamic]
    genreqs (Source _) p = pure [toDyn p]
    genreqs (Project v _) p = genreqs v p
    genreqs (Translate v tr) p = genreqs v (tr p)
    genreqs (Focus v _) p = (toDyn p :) <$> genreqs v (fst p)
    genreqs (Product vl vr) (p, q) = (++) <$> genreqs vl p <*> genreqs vr q
    genreqs (Merge vl vr _ _) p
      = (toDyn p :) <$> either (genreqs vl) (genreqs vr) p
    genreqs (Compose vg vf) pp@(MkComposed p) 
      = mrg pp <$> genreqs vg p <*> (genreqs vf =<< read vg p)
    mrg p ls rs = toDyn p : ls ++ rs
